#version 330 core

layout (std140) uniform Material {
	vec4 diffuse;
	vec4 ambient;
	vec4 specular;
	vec4 emissive;
	float shininess;
	int texCount;
} material;

struct Light {
	vec3 position;
	vec4 ambient;
	vec4 diffuse;
	vec4 specular;
	float constantAttenuation;
	float linearAttenuation;
	float quadraticAttenuation;
	vec3 direction;
	float spotCutoff;
	float spotExponent;
};

layout (std140) uniform Lights {
	Light light_array[2];
} lights;

uniform sampler1D toonUnit;
uniform sampler2D heightUnit;
uniform sampler2D shadowUnit;
uniform sampler2D texUnit;

uniform int drawOutline;
uniform int isBillboard;
uniform mat4 V;
uniform int enableToonShading;
uniform int shadowEnable;
uniform int width;
uniform int height;

in vec2 fTexCoord;
in vec3 fNormal;
in vec3 fViewDirection;
// in vec3 fPosition;
in vec3 fLightDir;
in vec3 fH;
in float fAttenuation;
in vec4 fShadowTexCoord;

out vec4 output;

float quantify(float val) {
	return enableToonShading == 1 ? texture(toonUnit, val).r : val;
}

float random(vec3 seed3, int idx) {
	vec4 seed4 = vec4(seed3, idx);
	float dot_product = dot(seed4, vec4(12.9898,78.233,45.164,94.673));
    return fract(sin(dot_product) * 43758.5453);
}

void main()
{
	vec4 color;
	vec4 amb;

	if (material.texCount == 0) {
		color = material.diffuse;
		amb = material.ambient;
	} else {
		color = texture2D(texUnit, fTexCoord);
		amb = color * 0.33;
		
		if (color.a == 0) {
			discard;
		}

		if (material.diffuse.rgb == vec3(0.0)) {
			output = color;
			return;
		}
		
		if (isBillboard == 1) {
			amb = color * 0.5;
		}
	}

	vec3 normal = normalize(fNormal);
	vec3 lightDir = normalize(fLightDir);
	vec3 h = normalize(fH);

	float diffuseIntensity = max(0.0, dot(lightDir, normal));
	float specularIntensity = max(0.0, dot(normal, h));

	output.rgb = (lights.light_array[0].diffuse.rgb * color.rgb * quantify(diffuseIntensity * fAttenuation) + lights.light_array[0].ambient.rgb * amb.rgb);
	if (diffuseIntensity > 0.0) {
		float fSpec = pow(specularIntensity, material.shininess);
		output.rgb += lights.light_array[0].specular.rgb * material.specular.rgb * (quantify(fAttenuation * fSpec * diffuseIntensity));
	}
	output.a = color.a;

	if (shadowEnable == 0) return;
	// shadow
	vec2 smCoords = fShadowTexCoord.xy / fShadowTexCoord.w;
	if (any(lessThan(smCoords.xy, vec2(0.0))) || any(greaterThan(smCoords.xy, vec2(1.0)))) {
		return;
	}
	vec2 poissonDisk[4] = vec2[](
	  vec2( -0.94201624, -0.39906216 ),
	  vec2( 0.94558609, -0.76890725 ),
	  vec2( -0.094184101, -0.92938870 ),
	  vec2( 0.34495938, 0.29387760 )
	);

	float visibility = 1.0;
	float bias = 0.005;
	for (int i = 0; i < 4; ++i) {
		//int index = int(16.0 * random(gl_FragCoord.xyy,i)) % 4;
		//visibility -= 0.2 * (1.0 - texture2D(shadowUnit, vec3(fShadowTexCoord.xy + poissonDisk[index] * vec2(1/width, 1/height)) / fShadowTexCoord.w, (fShadowTexCoord.z - bias) / fShadowTexCoord.w)));
		float result = texture2D(shadowUnit, vec2(fShadowTexCoord.xy + poissonDisk[i] / 700) / fShadowTexCoord.w).z;
		if (result < ((fShadowTexCoord.z - bias) / fShadowTexCoord.w)) {
			visibility -= 0.15 * (1.0 - result);
		}
	}
	
	
	/*
	float shadow = textureProj(shadowUnit, fShadowTexCoord);
	output.rgb *= min(1.0, (shadow+0.5));*/
	output.rgb *= clamp(visibility, 0.0, 1.0);
	/*
	if (shadow == 0.0f) {
		output.rgb *= 0.5;
	}*/
}
